module Enviso::Authentication
	require 'digest'
  require 'base64'
  require 'openssl'
  require 'jwt'

  # The communication layer implements all the methods available in the Enviso API
  class << self
    attr_accessor :auth_token
    attr_accessor :refresh_token

    # Set's the default value's to nil and false
    # @return [Hash] configuration options
    def init!
      @defaults = {
        :@auth_token    => nil,
        :@refresh_token => nil
      }
    end

  	
    # in the authentication documentation, chapter 4.1:
    # 1.  Concatenate the API key with the current timestamp in the format below:
    #     <<APIKEY>>_<<timestamp(yyyy'-'MM'-'ddTHH:mm:ss.fffZ)>>
    # NOTE: The timestamp is in ISO 8601 
    def timestamp_for_authentication
      return Time.now.utc.strftime('%Y-%m-%dT%H:%M:%S.%LZ')
    end
  	
    # The method calculates the login signature, as described
		# 2.  Create a SHA256 hash of this value. Based on the hashed value, a digital signature will be created.

  	def calculate_login_signature(timestamp)
  		api_key = Enviso::Config.api_key
  		return Digest::SHA256.hexdigest("#{api_key}_#{timestamp}")
      # return OpenSSL::Digest::SHA256.new("#{api_key}_#{timestamp}")
  	end


    # Signs the login signature hash
		# From the authentication documentation, chapter 4.1:
		# 3. The combination of the created signature along with the provided API secret key will act as the digital signature of the call.

  	def sign_login_signature(key)
      secret = Enviso::Config.api_secret
      public_key = OpenSSL::PKey::RSA.new(secret)
      encrypted_string = Base64.encode64(public_key.public_encrypt(key))
      return encrypted_string
  	end

    # This method combines all the login steps
    # 1. generate a key (SHA256)
    # 2. signs the key with public_key
    # 3. sends request to API endpoint

    def login(signed_key: nil)
      self.auth_token  = nil # reset to prevent giving auth token in login request headers
      
      timestamp   = timestamp_for_authentication
      key         = calculate_login_signature(timestamp)
      signed_key  = sign_login_signature(key) unless signed_key

      body = {
        apikey: Enviso::Config.api_key,
        timestamp: timestamp,
        signature: signed_key
      }
      result = Enviso::API.send_request(type: :post, endpoint: "apis/login", body: body)
      if result["authToken"]
        self.refresh_token = result["refreshToken"]
        self.auth_token = result["authToken"]
      end
      return result
    end

    # Get new access token using refresh token from login request response
    def get_new_token
      body = { refreshToken: refresh_token }
      puts "Refreshing token" if Enviso::Config.verbose
      begin
        result = Enviso::API.send_request(type: :post, endpoint: "apis/getnewtoken", body: body)
        if result["token"]
          self.auth_token = result["token"]
        end
        return result
      rescue
        puts "Signing in again to refresh token" if Enviso::Config.verbose
        login
      end
    end

    # Parses the current API key and checks if the 'exp' attribute
    # represents a date in the future.
    def has_valid_api_key
      return self.auth_token != nil && api_key_valid_until > Time.now
    end

    def api_key_valid_until
      decoded_token = JWT.decode self.auth_token, nil, false
      Time.at(decoded_token.first["exp"])
    end

  end
  
  init!
end